#ifndef _Binary_tree
#define _Binary_tree

#include <cassert>
#include <cstddef> // 'NULL'
#include <cstdlib> // 'int abs(int)'
#include <typeinfo>

template<class T> class Binary_tree {
protected:
	T _content;
	Binary_tree<T>* _left, * _right, * _super;
	virtual void left_binding(Binary_tree<T>*, Binary_tree<T>*) const;
	virtual void left_unbinding(Binary_tree<T>*, Binary_tree<T>*) const;
	virtual void right_binding(Binary_tree<T>*, Binary_tree<T>*) const;
	virtual void right_unbinding(Binary_tree<T>*, Binary_tree<T>*) const;
public:
	Binary_tree(const T&);
	int h() const;
	int H() const;
	virtual bool isAVL() const;
	virtual void push_left(Binary_tree<T>*);
	virtual void push_right(Binary_tree<T>*);

	inline const Binary_tree<T>* root() const {
		if (!_super) return this;
		return _super->root();
	};

	inline Binary_tree<T>*& super() {
		return _super;
	};
};

template<class T> void Binary_tree<T>::left_binding(Binary_tree<T>* down, Binary_tree<T>* up) const {
	down->_super = up;
	up->_left = down;
}

template<class T> void Binary_tree<T>::left_unbinding(Binary_tree<T>* down, Binary_tree<T>* up) const {
	down->_super = NULL;
	up->_left = NULL;
}

template<class T> void Binary_tree<T>::right_binding(Binary_tree<T>* down, Binary_tree<T>* up) const {
	down->_super = up;
	up->_right = down;
}

template<class T> void Binary_tree<T>::right_unbinding(Binary_tree<T>* down, Binary_tree<T>* up) const {
	down->_super = NULL;
	up->_right = NULL;
}

template<class T> Binary_tree<T>::Binary_tree(const T& t) : _content(t) {
	_left = _right = _super = NULL;
}

template<class T> int Binary_tree<T>::h() const {
	if (!_super) return 0;
	return _super->h() + 1;
}

template<class T> int Binary_tree<T>::H() const {
	int H_left, H_right;
	_left ? H_left = _left->H() + 1 : H_left = 0;
	_right ? H_right = _right->H() + 1 : H_right = 0;
	return H_left < H_right ? H_right : H_left;
}

template<class T> bool Binary_tree<T>::isAVL() const {
	if (!_left && !_right) return true;
	if (!_left) return _right->H() < 1;
	if (!_right) return _left->H() < 1;
	return ::abs(_right->H() - _left->H()) <= 1;
}

template<class T> void Binary_tree<T>::push_left(Binary_tree<T>* left) {
	assert(left);
	if (_left || left->_super) return;
	left_binding(left, this);
}

template<class T> void Binary_tree<T>::push_right(Binary_tree<T>* right) {
	assert(right);
	if (_right || right->_super) return;
	right_binding(right, this);
}

template<class T> class AVL_tree : public Binary_tree<T> {
protected:
	virtual void left_binding(Binary_tree<T>*, Binary_tree<T>*) const;
	virtual void right_binding(Binary_tree<T>*, Binary_tree<T>*) const;
public:
	AVL_tree(const T&);
	virtual bool isAVL() const;
	virtual void push_left(Binary_tree<T>*);
	virtual void push_right(Binary_tree<T>*);
};

template<class T> void AVL_tree<T>::left_binding(Binary_tree<T>* down, Binary_tree<T>* up) const {
	Binary_tree<T>::left_binding(down, up);
	const Binary_tree<T>* root = this->root();
	if (typeid (*root) == typeid (const AVL_tree<T>) && !root->Binary_tree<T>::isAVL()) left_unbinding(down, up);
}

template<class T> void AVL_tree<T>::right_binding(Binary_tree<T>* down, Binary_tree<T>* up) const {
	Binary_tree<T>::right_binding(down, up);
	const Binary_tree<T>* root = this->root();
	if (typeid (*root) == typeid (const AVL_tree<T>) && !root->Binary_tree<T>::isAVL()) right_unbinding(down, up);
}

template<class T> AVL_tree<T>::AVL_tree(const T& t) : Binary_tree<T>(t) {
}

template<class T> bool AVL_tree<T>::isAVL() const {
	return true;
}

template<class T> void AVL_tree<T>::push_left(Binary_tree<T>* left) {
	assert(left);
	if (_left || left->super()) return;
	if (!_right && left->H() >= 1) return;
	if (_right && ::abs(_right->H() - left->H()) > 1) return;
	left_binding(left, this);
}

template<class T> void AVL_tree<T>::push_right(Binary_tree<T>* right) {
	assert(right);
	if (_right || right->super()) return;
	if (!_left && right->H() >= 1) return;
	if (_left && ::abs(_left->H() - right->H()) > 1) return;
	right_binding(right, this);
}

#endif
